program sudko

  use orderedList, testOl => test

  implicit none

  integer, parameter :: REC  = b'000001' ! Reconcile board       (1)
  integer, parameter :: FA_S = b'000010' ! Frequency analysis    (2)
  integer, parameter :: SG   = b'000100' ! Subgrid shadow        (4)
  integer, parameter :: RR   = b'001000' ! Ring reduction        (8) 
  integer, parameter :: PH   = b'010000' ! Pidgeonhole reduction (16)
  integer, parameter :: WI   = b'100001' ! What-if               (33)

  integer, parameter :: RM_SINGLE = 1    ! Run a single step - do not loop
  integer, parameter :: RM_FINAL  = 2    ! Only output the final result
  integer, parameter :: RM_GENERATE = 3  ! Only output the final result

  integer, parameter :: FH_METHODS = 72  ! File unit number for methods file

  integer :: i, j, k, m_i, m_j, m_z, iter, num_changes, ccount
  integer :: lm_i, lm_j, lm_z

  type(oList), dimension(:,:), allocatable :: board, sboard
  integer            :: sg_rows, sg_columns
  logical            :: solved, updated, error
  integer, parameter :: fh = 70
  character(len=3)   :: arg
  integer            :: solveMode, runMode, s, phSize

  ! Read the solve mode from the arg list
  runMode = 0
  solveMode = 0
  phSize = 2
  s = command_argument_count()
  if (s >= 1) then
     call get_command_argument(1,arg)
     arg = trim(arg)
     read(arg,'(i3)') solveMode
     if (s >= 2) then
        call get_command_argument(2,arg)
        read(arg,'(i3)') runMode
        if (s >= 3) then
           call get_command_argument(3,arg)
           read(arg,'(i2)') phSize
        end if
     end if
  end if
  write(fh,'(3(a,i0))') 'solveMode = ',solveMode,'; runMode = ',runMode,'; phSize = ',phSize

  ! Run Olist test if requested
  if (runMode == 10) then
     call testOl
     stop
  end if

  ! Generate board if requested
  if (runMode == RM_GENERATE) then
     stop
  end if

  ! Load the puzzle board
  call loadBoard

  ! Initial output
  call print(70)
  if (runMode /= 2) then
     call write
  end if

  ! Initial reconcile of the board to check it's valid
  sboard = board
  updated = reconcile(m_i, m_j, m_z, error)
  if (error) then
     write(*,'(a)') '***Invalid board'
     stop
  end if
  board = sboard

  ! Open the methods file
  open(72,file='methods.tmp',status='replace')

  ! Now process the data using the specified methods
  iter = 0
  call update(m_i, m_j, m_z, error)

  ! If the board still hasn't been solved and What-if? has been requested
  ! then do that
  if (m_z > 1 .and. useMethod(WI)) then
     call whatIf(m_i, m_j, m_z)
  end if

  ! All done
  call print(70)
  if (runMode == RM_FINAL) then
     call write
  end if
  close(72)
  stop

contains

  ! ---------------------------------------------------------------------------

  recursive subroutine whatIf(m_i, m_j, m_z)
    integer, intent(inout) :: m_i, m_j
    integer, intent(inout) :: m_z

    integer :: k
    type(oList), dimension(:,:), allocatable :: sboard
    if (m_z == 1) return
    allocate(sboard(size(board,1),size(board,2)))
    sboard = board
    error = .false.
    do k=1,sboard(m_i,m_j)%size
       board = sboard
       board(m_i,m_j)%size = 1
       board(m_i,m_j)%data = sboard(m_i,m_j)%data(k)
       if (error) then
          call write(WI)
       end if
       call update(m_i, m_j, m_z, error)
       if (error) cycle
       if (m_z > 1) then
          call whatIf(m_i, m_j, m_z)
       end if
       if (m_z == 1) exit
    end do
    deallocate(sboard)
  end subroutine whatIf

  ! ---------------------------------------------------------------------------

  subroutine update(m_i, m_j, m_z, error)
    integer, intent(out) :: m_i, m_j, m_z
    logical, intent(out) :: error
    logical :: updated, updated_fa, updated_rg, updated_sg
    integer :: phole
    logical, allocatable :: updated_ph(:)

    allocate(updated_ph(ccount-3))
    error = .false.
    m_z = 1
    do
       if (useMethod(REC)) then
          updated = reconcile(m_i, m_j, m_z, error)
          if (updated) call write(REC)
          if (error .or. m_z == 1) exit
       end if
       iter = iter + 1
       if (useMethod(FA_S)) then
          updated_fa = frequencyAnalysis()
          if (updated_fa) then
             call write(FA_S)
             if (useMethod(REC)) then
                updated = reconcile(m_i, m_j, m_z, error)
                if (updated) call write(REC)
                if (error .or. m_z == 1) exit
             end if
          end if
       else
          updated_fa = .false.
       end if

       if (useMethod(SG)) then
          updated_sg = sgridShadow()
          if (updated_sg) then
             call write(SG)
             if (useMethod(REC)) then
                updated = reconcile(m_i, m_j, m_z, error)
                if (updated) call write(REC)
                if (error .or. m_z == 1) exit
             end if
          end if
       else
          updated_sg = .false.
       end if

       if (useMethod(RR)) then
          updated_rg = ringReduction()
          if (updated_rg) then
             call write(RR)
             if (useMethod(REC)) then
                updated = reconcile(m_i, m_j, m_z, error)
                if (updated) call write(REC)
                if (error .or. m_z == 1) exit
             end if
          end if
       else
          updated_rg = .false.
       end if

       updated_ph = [(.false.,phole=1,size(updated_ph))]
       if (useMethod(PH)) then
          do phole = 2,min(phSize,ccount-2)
             updated_ph(phole-1) = pidgeonHoles(phole)
             if (updated_ph(phole-1)) then
                call write(PH,ph_size=phole)
                if (useMethod(REC)) then
                   updated = reconcile(m_i, m_j, m_z, error)
                   if (updated) call write(REC)
                   if (error .or. m_z == 1) exit
                end if
             end if
          end do
       end if
       if ((.not. updated_fa .and. .not. updated_sg .and. .not. updated_rg .and. &
            count(updated_ph)==0) .or. runMode == RM_SINGLE) exit
    end do


  end subroutine update

  ! ---------------------------------------------------------------------------
  logical function pidgeonHoles(n)
    integer, intent(in) :: n
    ! Look for n numbers distributed between n cells within a row, column
    ! or subgrid
    integer              :: i, j, k, l, m, si, sj, mode, old, s, e
    type(oList)          :: c
    logical, allocatable :: used_ph(:)
    logical              :: found
    integer              :: cc(2)

    pidgeonHoles = .false.
    allocate(used_ph(ccount))

    ! Process the rows and columns
    do mode=0,1
       do i=1,size(board,1)
          if (runMode > 4) write(*,'(2(a,i0))') '[PH',n,']: checking row ',i
          ! Initialize data for row scan
          do j=1,size(board,2)-n+1
             if (board(i,j)%size == 1 .or. board(i,j)%size > n) cycle
             used_ph = spread(.false.,1,size(used_ph))
             c = board(i,j)
             used_ph(j) = .true.
             if (runMode > 4 .and. debug(i)) write(*,'(a,i0)') 'starting at column ',j
             found = addPhSlot(n, used_ph, c, i, j+1)
             if (debug(i)) write(*,'(a,l1)') ':--> found = ',found
             if (found) then            
                if (runMode > 4) write(*,'(2(a,i0))') '[PH',n,'] match in row ',i
                do k=1,size(board,2)
                   if (board(i,k)%size > 1 .and. .not. used_ph(k)) then
                      if (runMode > 4) write(*,*) '[PH',n,'] removing ',(c%data(l),l=1,c%size),' from cell ',k
                      old = board(i,k)%size
                      board(i,k) = board(i,k) - c
                      pidgeonHoles = pidgeonHoles .or. (board(i,k)%size < old)
                   end if
                end do
                exit
             end if
          end do
       end do
       board = transpose(board)
    end do

    ! Process the subgrids
    do si=1,sg_rows
       do sj=1,sg_columns
          do m=1,ccount-n+1
             call getSubgridCell(si,sj,m,i,j)
             if (board(i,j)%size == 1 .or. board(i,j)%size > n) cycle
             used_ph = spread(.false.,1,size(used_ph))
             c = board(i,j)
             used_ph(m) = .true.
             found = addPhSgSlot(n, si, sj, used_ph, c, m+1)
             if (found) then
                do k=1,size(board,2)
                   call getSubgridCell(si,sj,k,i,j)
                   if (board(i,j)%size > 1 .and. .not. used_ph(k)) then
                      old = board(i,k)%size
                      board(i,j) = board(i,j) - c
                      pidgeonHoles = pidgeonHoles .or. (board(i,k)%size < old)
                   end if
                end do
                exit
             end if
          end do
       end do
    end do

  end function pidgeonHoles

  recursive function addPhSlot(n, used_ph, c, i, j) result(aph)
    logical :: aph
    integer, intent(in) :: n
    logical, intent(inout) :: used_ph(:)
    type(oList), intent(inout) :: c
    integer, intent(in) :: i, j
    type(oList) :: temp
    integer :: k

    aph = .true.
    if (j > size(used_ph)) then
       ! No more to try so return true/false according to the current state
       aph = (c%size == n .and. count(used_ph) == n)
       return
    end if
    if (board(i,j)%size == 1 .or. board(i,j)%size > n) then
       ! Cell has too many numbers or is already fixed - try the next one
       aph = addPhSlot(n, used_ph, c, i, j+1)
       return
    end if
    temp = c + board(i,j)
    if (temp%size > n) then
       ! If this cell were added the number of entries in our list would exceed
       ! the number of pidgeonholes
       aph = addPhSlot(n, used_ph, c, i, j+1)
       return
    end if
    if (count(used_ph) == n) then
       ! This is an error (i.e. a board inconsistency)
       aph = .false.
       return
    end if
    ! The temp list is the new current list so mark the pidgeonhole as
    ! used and move to the next one.
    c = temp
    used_ph(j) = .true.
    aph = addPhSlot(n, used_ph, c, i, j+1)
  end function addPhSlot

  recursive function addPhSgSlot(n, si, sj, used_ph, c, m) result(aph)
    logical :: aph
    integer, intent(in) :: n, si, sj
    logical, intent(inout) :: used_ph(:)
    type(oList), intent(inout) :: c
    integer, intent(in) :: m
    type(oList) :: temp
    integer :: k


    call getSubgridCell(si, sj, m, i, j)
    aph = .true.
    if (m > size(used_ph)) then
       aph = (c%size == n .and. count(used_ph) == n)
       return
    end if
    if (board(i,j)%size == 1 .or. board(i,j)%size > n) then
       aph = addPhSgSlot(n, si, sj, used_ph, c, m+1)
       return
    end if
    temp = c + board(i,j)
    if (temp%size > n) then
       aph = addPhSgSlot(n, si, sj, used_ph, c, m+1)
       return
    end if
    if (count(used_ph) == n) then
       aph = .false.
       return
    end if
    c = temp
    used_ph(m) = .true.
    aph = addPhSgSlot(n, si, sj, used_ph, c, m+1)
  end function addPhSgSlot

  ! ---------------------------------------------------------------------------
  subroutine getSubgridCell(si, sj, m, i, j)
    integer, intent(in) :: si, sj, m
    integer, intent(inout) :: i, j
    
    i = (si-1)*sg_columns+1 + (m-1)/sg_rows
    j = (sj-1)*sg_rows+1 + mod(m-1,sg_rows)

  end subroutine getSubgridCell
  ! ---------------------------------------------------------------------------

  logical function frequencyAnalysis()
    integer, dimension(:), allocatable :: freq
    integer, dimension(:), allocatable :: pairFreq
    logical, dimension(:), allocatable :: havePair
    integer, dimension(:), allocatable :: pairs
    integer                            :: i, j, si, sj, k, mode

    frequencyAnalysis = .false.

    allocate(freq(size(board,1)))

    ! Frequency constraints - rows
    write(fh,'(a)') 'doing frequency analysis:'
    do mode=0,1
       do i=1,size(board,1)
          freq = spread(1+size(freq),1,size(freq))
          do j=1,size(board,2)
             call cumFreq(freq,i,j)
          end do
          if (minval(freq) == 1) then
             write(fh,'(a,i0,a)',advance='no') 'FA_S: row ',i,': => '
             frequencyAnalysis = frequencyAnalysis .or. updateRow(i,minloc(freq,1))
          end if
       end do
       if (frequencyAnalysis) then
          if (mode == 1)  board = transpose(board) ! Switch back before returning
          return
       end if
       board = transpose(board)
    end do

    ! Frequency constraints - subgrids
    do si=1,sg_rows
       do sj=1,sg_columns
          freq = spread(1+size(freq),1,size(freq))
          do k=1,ccount
             call getSubgridCell(si,sj,k,i,j)
             call cumFreq(freq,i,j)
          end do
          if (minval(freq) == 1) then
             write(fh,'(2(a,i0),a)',advance='no') 'FA_S: subgrid [',si,',',sj,']: => '
             frequencyAnalysis =  frequencyAnalysis .or. &
                  updateSgrid(si,sj,minloc(freq,1))
          end if
        end do
    end do

  end function frequencyAnalysis

  subroutine cumFreq(freq,row,col)
    integer, dimension(*), intent(inout) :: freq
    integer, intent(in)                  :: row, col
    integer :: k
    if (board(row,col)%size == 1) return
    do k=1,board(row,col)%size
       if (freq(board(row,col)%data(k)) == 10) then
          freq(board(row,col)%data(k)) = 1
       else
          freq(board(row,col)%data(k)) = freq(board(row,col)%data(k)) + 1
       end if
    end do
  end subroutine cumFreq

  logical function updateRow(row, value)
    integer, intent(in) :: row, value
    integer :: j, k
    updateRow = .false.
    do j=1,size(board,2)
       if (board(row,j)%size == 1) cycle
       do k=1,board(row,j)%size
          if (board(row,j)%data(k) == value) then
             board(row,j)%size = 1
             board(row,j)%data(1) = value
             write(fh,'(3(a,i0))') 'cell (',row,',',j,') set to ',value
             updateRow = .true.
             return
          end if
       end do
    end do
  end function updateRow

  logical function updateSgrid(si, sj, value)
    integer, intent(in) :: si, sj, value
    integer :: i, k, m

    updateSgrid = .false.
    do m=1,ccount
       call getSubgridCell(si,sj,m,i,j)
       if (board(i,j)%size == 1) cycle
       do k=1,board(i,j)%size
          if (board(i,j)%data(k) == value) then
             board(i,j)%size = 1
             board(i,j)%data(1) = value
             write(fh,'(3(a,i0))') 'cell (',i,',',j,') set to ',value
             updateSgrid = .true.
             return
          end if
       end do
    end do
  end function updateSgrid

  ! ---------------------------------------------------------------------------
  ! Subgrid shadow analysis
  !
  ! Look for frequencies of 2 or 3 within a row or column.
  logical function sgridShadow()
    integer, allocatable  :: freq(:)
    integer               :: si, sj, pcount, l, old, m
    integer, dimension(2) :: o
    type lp
       integer, dimension(2) :: p
    end type lp
    type(lp), dimension(3) :: q
    logical                :: aligned, ud

    sgridShadow = .false.
    allocate(freq(size(board,1)))
    do si=1,sg_rows
       do sj=1,sg_columns

          freq = spread(1+size(freq),1,size(freq))
          do m=1,ccount
             call getSubgridCell(si,sj,m,i,j)
             call cumFreq(freq,i,j)
          end do

          ! Find the coordinates of each freq 2 & 3 number
          o = [(si-1)*sg_rows+1,(sj-1)*sg_columns+1]
          do k=1,size(freq)
             if (freq(k) /= 2  .and. freq(k) /= 3) cycle
             pcount = 0
             do i=(si-1)*sg_columns+1,si*sg_columns
                do j=(sj-1)*sg_rows+1,sj*sg_rows
                   if (board(i,j) .c. k) then
                      pcount = pcount + 1
                      q(pcount)%p = [i,j]
                   end if
                end do
             end do
             if (q(2)%p(1) == q(1)%p(1)) then
                o = [q(2)%p(1), 0]
                aligned = .true.
             elseif (q(2)%p(2) == q(1)%p(2)) then
                o = [0, q(1)%p(2)]
                aligned = .true.
             else
                o = [0,0]
                aligned = .false.
             end if
             if (aligned .and. pcount == 3) then
                if (q(3)%p(1) /= o(1) .and.q(3)%p(2) /= o(2) ) then
                   aligned = .false.
                end if
             end if

             if (aligned) then
                ud = .false.
                if (o(1) /= 0) then
                   do j=1,size(board,2)
                      if (j >= (sj-1)*sg_columns+1 .and. j <= sj*sg_columns) cycle
                      old = board(o(1),j)%size
                      board(o(1),j) = board(o(1),j) - k
                      ud = ud .or. (old > board(o(1),j)%size)
                   end do
                   if (ud) then
                      write(fh,'(2(a,i0),a,2i2)') 'removed ',k,' from row ',o(1), &
                           ' outside sub-grid ',si,sj
                      sgridShadow = .true.
                   end if
                else
                   do i=1,size(board,1)
                      if (i >= (si-1)*sg_rows+1 .and. i <= si*sg_rows) cycle
                      old = board(i,o(2))%size
                      board(i,o(2)) = board(i,o(2)) - k
                      ud = ud .or. (old > board(i,o(2))%size)
                   end do
                   if (ud) then
                      write(fh,'(2(a,i0),a,2i2)') 'removing ',k,' from column ', &
                           o(2),' outside sub-grid ',si,sj
                      sgridShadow = .true.
                   end if
                end if
             end if
          end do
       end do
    end do
    deallocate(freq)
  end function sgridShadow

  ! ---------------------------------------------------------------------------

  ! Ring reduction
  logical function ringReduction()
    type(oList), dimension(4) :: nodes
    integer                   :: node_ind, i, j, k, l
    integer, pointer          :: x
    type(oList), target       :: d1, d2, d3, r
    ! Define type for the logical positin on a board
    type lpos
       integer, dimension(2) :: p
    end type lpos
    type(lpos), dimension(4), target :: coords
    type(lpos), pointer :: top_left, top_right, bottom_left, bottom_right


    top_left    => coords(1); top_right    => coords(2);
    bottom_left => coords(3); bottom_right => coords(4);

    ! Search for a ring
    write(fh,'(a)') 'doing ring reduction:'
    ringReduction = .false.
    row: do i=1,size(board,1)
       node_ind = 0
       col: do j=1,size(board,2)
          if (board(i,j)%size /= 2) cycle col
          select case (node_ind)
          case (0)
             node_ind = node_ind + 1
             top_left%p = [i,j]
          case (1)
             if (i == top_left%p(1)) then
                node_ind = node_ind + 1
                top_right%p = [i,j]
             else
                node_ind = 1
                top_left%p = [i,j]
             end if
          end select
          if (node_ind <2) cycle col

          level: do k=i+1,size(board,1)
             if (board(k,top_left%p(2))%size /= 2 .or. &
                  board(k,top_right%p(2))%size /= 2) cycle level
             bottom_left%p  = [k,top_left%p(2)]
             bottom_right%p = [k,top_right%p(2)]
             node_ind = 4
             exit level
          end do level
          if (node_ind /= 4) cycle col

          ! Check to see if this is a valid ring; (a,x) -> (b,x)
          !                                         |        |
          !                                       (c,x) -> (b,c)
          write(fh,'(a,(4(a,i0,a,i0,a)))') 'ring detected at coords: ', &
               ('-> (',coords(k)%p(1),',',coords(k)%p(2),') ',k=1,4)
          ! Allocate the data to the nodes
          do k=1,4
             nodes(k) = board(coords(k)%p(1),coords(k)%p(2))
          end do
          write(fh,'(a,(4(a,i0,a,i0,a)))') '            ring values: ', &
               ('-> (',nodes(k)%data(1),',',nodes(k)%data(2),') ',k=1,4)


          !     (a,x) -> (b,x)                 (a)  -> (b,x)
          !       |        |     ========>      |        |
          !       v        v                    v        v
          !     (c,x) -> (b,c)                (c,x) -> (b,c)

          do k=2,4,2
             l = mod(k,4) + 1
             d1 = nodes(k) - nodes(l) ! 2 - 3 or 4 - 1
             d2 = nodes(l) - nodes(k) ! 3 - 2 or 4 - 1
             r = d1 + d2
             d3 = nodes(2) - r
             x => d3%data(1)
             if (r == nodes(k-1) .and. (nodes(l+1) .C. x)) then
                write(fh,'(3(a,i0),a)') 'elimating ',x,' from node at (', &
                     coords(l+1)%p(1),',',coords(l+1)%p(2),')'
                board(coords(l+1)%p(1),coords(l+1)%p(2)) = &
                     board(coords(l+1)%p(1),coords(l+1)%p(2)) - x
                ringReduction = .true.
                cycle row
                
             elseif (r == nodes(l+1) .and. (nodes(k-1) .C. x)) then
                write(fh,'(3(a,i0),a)') 'elimating ',x,' from node at (', &
                     coords(k-1)%p(1),',',coords(k-1)%p(2),')'
                board(coords(k-1)%p(1),coords(k-1)%p(2)) = &
                     board(coords(k-1)%p(1),coords(k-1)%p(2)) - x
                ringReduction = .true.
                cycle row
             endif
          end do
          write(fh,'(a)') 'no reduction possible'
       end do col
    end do row

  end function ringReduction

  ! ---------------------------------------------------------------------------


  integer function updateEntry(row, col)
    integer, intent(in)    :: row, col
    integer :: i,j, old, rcol, rrow, rsgr, si, sj, k, m
    type(oList) :: constraint
    updateEntry = 0

    if (board(row,col)%size == 1) return ! Nothing to do

    if (debug(row,col)) then
       write(fh,'(a,i0,a,i0,a)') 'updating cell (',row,',',col,')'
       write(fh,*) 'current value: ',(board(row,col)%data(k),k=1,board(row,col)%size)
    end if

    ! Constraints from the row
    r: do j=1,size(board,2)
       if (j /= col .and. board(row,j)%size == 1) then
          constraint = constraint + board(row,j)%data(1)
       end if
    end do r
    if (debug(row,col)) write(fh,*) (constraint%data(k),k=1,constraint%size)

    ! Constraints from the column
    c: do i=1,size(board,1)
       if (i /= row .and. board(i,col)%size == 1) then
          constraint = constraint + board(i,col)%data(1)
       end if
    end do c
    if (debug(row,col)) write(fh,*) (constraint%data(k),k=1,constraint%size)

    ! Constraints from the subgrid
    si = 1+(row-1)/sg_columns
    sj = 1+(col-1)/sg_rows
    do m=1,ccount
       call getSubgridCell(si,sj,m,i,j)
       if (i /= row .and. j /= col .and. board(i,j)%size == 1) then
          constraint = constraint + board(i,j)%data(1)
       end if
    end do
    if (debug(row,col)) write(fh,*) (constraint%data(k),k=1,constraint%size)

    old = board(row,col)%size
    board(row,col) = board(row,col) - constraint
    updateEntry = old - board(row,col)%size

  end function updateEntry

  subroutine print(fh)
    integer, optional, intent(in) :: fh
    integer :: i, j, k, w
    integer :: ounit
    integer, allocatable :: cwidth(:)
    allocate(cwidth(size(board,1)))

    if (present(fh)) then
       ounit = fh
    else
       ounit = 6
    end if

    ! Find max col width
    w = 0
    do j=1,size(board,2)
       k = 0
       do i=1,size(board,1)
          if (board(i,j)%size /= size(board,1)) then
             k = max(k,2*board(i,j)%size-1)
          end if
       end do
       cwidth(j) = k+2
       if (k == 1) then
          w = w + 8
       else
          w = w + cwidth(j) + 4
       end if
    end do
    ! Output the board
    do i=1,size(board,1)
       if (mod(i-1,sg_columns) == 0) then
          write(ounit,'(a)') repeat('- ',1+w/2)
       end if
       do j=1,size(board,2)
          if (board(i,j)%size == 1) then
             if (mod(j-1,sg_rows) == 0) then
                write(ounit,'(a,i0)',advance='no') '|   ',board(i,j)%data(1)
             else
                write(ounit,'(a,i0)',advance='no') ':   ',board(i,j)%data(1)
             end if
             if (cwidth(j) > 4) then
                write(ounit,'(a)',advance='no') repeat(' ',cwidth(j)-4)
             end if
             write(ounit,'(a)',advance='no') '   '

          elseif (board(i,j)%size == size(board,1)) then
             if (mod(j-1,sg_rows) == 0) then
                write(ounit,'(a,i0)',advance='no') '|   ',0
             else
                write(ounit,'(a,i0)',advance='no') ':   ',0
             end if
             if (cwidth(j) > 4) then
                write(ounit,'(a)',advance='no') repeat(' ',cwidth(j)-4)
             end if
             write(ounit,'(a)',advance='no') '   '

          else
             if (mod(j-1,sg_rows) == 0) then
                write(ounit,'(a)',advance='no') '| ('
             else
                write(ounit,'(a)',advance='no') ': ('
             end if
             do k=1,board(i,j)%size-1
                write(ounit,'(i0,a)',advance='no') board(i,j)%data(k),','
             end do
             write(ounit,'(i0,a)',advance='no') board(i,j)%data(board(i,j)%size),')'
             if (cwidth(j) > 2*board(i,j)%size-1) then
                write(ounit,'(a)',advance='no') repeat(' ', &
                     cwidth(j)-(2*board(i,j)%size-1))
             end if
          end if
       end do
       write(ounit,'(a)') '|'
    end do
    write(ounit,'(a)') repeat('- ',1+w/2)
    deallocate(cwidth)
  end subroutine print

  subroutine loadBoard
    character(len=256) :: field, line
    character(len=2)   :: el
    integer :: current, next, fcount, cloc
    integer :: e_c, e_n, k, dum
    
    ! Get the board dimensions
    read(*,'(a)',end=100) line
    
    ! Comma separated pair
    cloc = scan(line,',')
    read(line(1:cloc-1),*) sg_columns
    read(line(cloc+1:len_trim(line)),*) sg_rows
    ccount = sg_rows*sg_columns
    
    ! Allocate memory for the whole board
    allocate(board(ccount,ccount))
    allocate(sboard(ccount,ccount))
    write(fh,'(2(a,i0))') 'allocated board size = ',size(board,1),' x ',size(board,2)

    ! Read format e.g.:  1,0,3+5+7,2+9,0,0,0,0,1+2+3
    i = 0
    lines: do
       line = ' '
       read(*,'(a)',end=100) line
       i = i + 1
       current = 1
       fcount = 1
       row: do
          next = index(trim(line(current:)),',')+current-1
          if (next < current) then
             next = len(trim(line)) + 1
          end if
          field = line(current:next-1)
          if (trim(field) == '0' .or. trim(field) == '*') then
             ! All values already set
             board(i,fcount)%data = [(j,j=1,size(board,1))]
             board(i,fcount)%size = size(board,1)

          elseif (scan(field,'+') == 0) then
             read(field,'(i2)') board(i,fcount)%data(1)
             board(i,fcount)%size = 1

          else
             e_c = 1
             k  = 1
             ele:  do
                e_n = index(trim(field(e_c:)),'+')+e_c-1
                if (e_n < e_c) then
                   e_n = len(trim(field)) + 1
                end if
                el = field(e_c:e_n-1)
                read(el,'(i2)') dum
                board(i,fcount) =  board(i,fcount) + dum
                k = k + 1
                if (e_n == len(trim(field)) + 1) exit ele
                e_c = e_n + 1
             end do ele
          end if
          fcount = fcount + 1
          current = next + 1
          if (next == len(trim(line)) + 1) exit row
       end do row
    end do lines

100 continue
  end subroutine loadBoard

  logical function reconcile(m_i, m_j, m_z, error)
    integer, intent(out) :: m_i, m_j, m_z
    logical, intent(out) :: error
    integer              :: i, j, k, iter, num_changes, ue, uc
    logical, allocatable :: uniq(:)

    i = 0
    k = 0
    iter = 0
    num_changes = 0
    error = .false.
    reconcile = .false.
    ! Loop until there are no more changes
    r: do
       iter = iter + 1
       i = mod(i,size(board,1)) + 1
       j = 0
       c: do
          j = j+1
          if (j > size(board,2)) exit c
          ue = updateEntry(i,j)
          reconcile = reconcile .or. (ue > 0)
          num_changes = num_changes + ue
          if (board(i,j)%size == 0) then
             write(fh,'(a,i0,a,i0,a)') '***Error: inconsistent solution at (', &
                  i,',',j,')'
             error = .true.
             return
          end if
       end do c
       if (i == size(board,1)) then
          if (num_changes == 0) then
             exit r
          else
             num_changes = 0
          end if
       end if
    end do r

    ! Scan the board for the smallest size that is greater than 1
    m_z = 1
    m_i = 0
    m_j = 0
    do i=1,size(board,1)
       do j = 1,size(board,2)
          if (board(i,j)%size > 1 .and. (board(i,j)%size < m_z .or. m_z == 1)) then
             m_z = board(i,j)%size
             m_i = i
             m_j = j
          end if
       end do
    end do

    ! Check that single entries in a row or column are unique
    allocate(uniq(size(board,1)))
    row_un: do i=1,size(board,1)
       uniq = spread(.false.,1,size(uniq))
       uc = 0
       do j = 1,size(board,2)
          if (board(i,j)%size == 1) then
             uc = uc + 1
             uniq(board(i,j)%data(1)) = .true.
          end if
       end do
       if (count(uniq) /= uc) then
          write(fh,'(a,i0,a)') 'Row ',i,' has non-unique entries'
          error = .true.
       end if
    end do row_un

    col_un: do j=1,size(board,1)
       uniq = spread(.false.,1,size(uniq))
       uc = 0
       do i = 1,size(board,2)
          if (board(i,j)%size == 1) then
             uc = uc + 1
             uniq(board(i,j)%data(1)) = .true.
          end if
       end do
       if (count(uniq) /= uc) then
          write(fh,'(a,i0,a)') 'Column ',j,' has non-unique entries'
          error = .true.
       end if
    end do col_un

    write(fh,'(/i0,a,i0/)') iter,' iterations: m_z = ',m_z
  end function reconcile

  subroutine write(meth, fh, ph_size)
    integer, optional, intent(in) :: meth
    integer, optional, intent(in) :: fh
    integer, optional, intent(in) :: ph_size
    ! Dump  the current state to sdtout
    integer :: ounit

    ! Do not output intermediate data in runMode 2
    if (runMode == 2 .and. present(meth)) then
       return
    end if

    ! Set th output channel
    if (present(fh)) then
       ounit = fh
    else
       ounit = 6 ! this is the usual value but really it's compiler dependent
    end if

    ! Output the board subgrid size
    write(ounit,'(i0,a,i0)') sg_columns,',',sg_rows
    row: do i=1,size(board,1)
       col: do j=1,size(board,2)
          if (board(i,j)%size == size(board,1)) then
             write(ounit,'(a)',advance='no') '*'

          elseif (board(i,j)%size == 0) then
             write(ounit,'(a)',advance='no') 'X'

          else
             write(ounit,'(i0)',advance='no') board(i,j)%data(1)
             do k=2,board(i,j)%size
                write(ounit,'(a1,i0)',advance='no') '+',board(i,j)%data(k)
             end do
          end if
          if (j == size(board,2)) then
             write(ounit,'(/)',advance='no')
          else
             write(ounit,'(a)',advance='no') ','
          end if
       end do col
    end do row

    ! Cutput the method used
    if (present(meth)) then
       select case(meth)
          case (RR)
          write(FH_METHODS,'(a)') 'Ring reduction'

          case (REC)
          write(FH_METHODS,'(a)') 'Reconcile'

          case (PH)
          write(FH_METHODS,'(a,i0)') 'Pidgeonhole-',ph_size

          case (FA_S)
          write(FH_METHODS,'(a)') 'Singleton'

          case (SG)
          write(FH_METHODS,'(a)') 'Subgrid shadow'

          case (WI)
          write(FH_METHODS,'(a)') 'Backtracked'

          case default
          write(FH_METHODS,'(a,i0)') '***Unknown method - ',meth

       end select
    end if
  end subroutine write

  logical function useMethod(method)
    integer, intent(in) :: method
    useMethod = (iand(solveMode,method) == method)
  end function useMethod

  logical function debug(i, j)
    integer, intent(in) :: i
    integer, intent(in), optional ::j

    debug = merge(.true.,.false.,(i == 3) .and. runMode > 4)

  end function debug

  ! Generate a board
  subroutine generate
    
  end subroutine generate
end program sudko
